shifts = {
    "GUILDS": 0,
    "GUILD_MEMBERS": 1,
    "GUILD_BANS": 2,
    "GUILD_EMOJIS": 3,
    "GUILD_INTEGRATIONS": 4,
    "GUILD_WEBHOOKS": 5,
    "GUILD_INVITES": 6,
    "GUILD_VOICE_STATES": 7,
    "GUILD_PRESENCES": 8,
    "GUILD_MESSAGES": 9,
    "GUILD_MESSAGE_REACTIONS": 10,
    "GUILD_MESSAGE_TYPING": 11,
    "DIRECT_MESSAGES": 12,
    "DIRECT_MESSAGE_REACTIONS": 13,
    "DIRECT_MESSAGE_TYPING": 14
}


class Intents:
    """ A class to easily construct intents integers """

    def __init__(self, intents: list):
        self.intents = 0

        if len(intents) == 0 or "ALL" in intents:
            intents = [
                "GUILDS",
                "GUILD_BANS",
                "GUILD_EMOJIS",
                "GUILD_INTEGRATIONS",
                "GUILD_WEBHOOKS",
                "GUILD_INVITES",
                "GUILD_VOICE_STATES",
                "GUILD_MESSAGES",
                "GUILD_MESSAGE_REACTIONS",
                "GUILD_MESSAGE_TYPING",
                "DIRECT_MESSAGES",
                "DIRECT_MESSAGE_REACTIONS",
                "DIRECT_MESSAGE_TYPING"
            ]

        for intent in intents:
            intent = intent.upper()
            if intent in shifts:
                self.intents += 1 << shifts[intent]
            else:
                raise ValueError(f"Invalid Intent: {intent}")


from random import choices

default_charset = "0123456789abcdef"

def generate_id_solid(length: int = 32, charset: str = default_charset) -> str:
    return "".join(choices(charset, k=length))

def generate_id_split(length: int = 24, charset: str = default_charset, split_pattern: list = [6,10,6]) -> str:
    chars = generate_id_solid(length, charset)
    output = ""
    total = 0
    for n in split_pattern:
        while n != 0:
            output += chars[total]
            total += 1
            n -= 1
        output += "-"
    return output[:-1]



import discord
from discord.errors import DiscordException
from discord.ext.commands import check


class MissingAnyPermissions(DiscordException):
    pass


def has_any_permissions(**perms):
    """A check function for discord.py to check if a user has any of the given permissions, not all of them."""

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError('Invalid permission(s): %s' % (', '.join(invalid)))

    def predicate(ctx):
        ch = ctx.channel
        permissions = ch.permissions_for(ctx.author)

        for perm, value in perms.items():
            if value == False:
                continue
            if getattr(permissions, perm):
                return True

        raise MissingAnyPermissions(f"Missing one of the following permissions: {', '.join(perms.keys())}")

    return check(predicate)

def has_any_guild_permissions(**perms):
    """A check function for discord.py to check if a user has any of the given guild permissions, not all of them."""

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError('Invalid permission(s): %s' % (', '.join(invalid)))

    def predicate(ctx):
        if not ctx.guild:
            raise ValueError("This check must be called from a guild command.")

        permissions = ctx.author.guild_permissions

        for perm, value in perms.items():
            if value == False:
                continue
            if getattr(permissions, perm):
                return True

        raise MissingAnyPermissions(f"Missing one of the following guild permissions: {', '.join(perms.keys())}")

    return check(predicate)


"""
A collection of random utils
"""

__title__ = "vcoutils"
__author__ = "vcokltfre"

from .leakybucket import LeakyBucket, LeakyBucketManager
from .extrachecks import has_any_permissions, has_any_guild_permissions
from .intents import Intents
from .genid import generate_id_solid, generate_id_split


import setuptools

MODULES = 4
CLASSES = 3
BUGFIX = 1

VERSION = f"{MODULES}.{CLASSES}.{BUGFIX}"

REQS = []

with open("README.md") as f:
    DESC = f.read()

setuptools.setup(
    name="vcoutils",
    author="vcokltfre",
    author_email="vcokltfre@gmail.com",
    version=VERSION,
    url="https://github.com/vcokltfre/vcoutils",
    packages=setuptools.find_packages(),
    description="A collection of utlities I have made or adapted for convenience.",
    long_description_content_type="text/markdown",
    long_description=DESC,
    install_requires=REQS
)


from time import time


class LeakyBucket:
    def __init__(self, timeout: int):
        self.timeout = timeout
        self.items = []

    def add(self) -> int:
        self.items.append((len(self.items), time()))
        return self.get()

    def get(self) -> int:
        items = []
        for item in self.items:
            if not item[1] + self.timeout < time():
                items.append(item)
        self.items = items
        return len(self.items)

    def __str__(self):
        return f"<LeakyBucket n={len(self.items)} t={self.timeout}>"

    def __repr__(self):
        return f"<LeakyBucket n={len(self.items)} t={self.timeout}>"


class LeakyBucketManager:
    def __init__(self, default_timeout: int = 60):
        self.timeout = default_timeout
        self.buckets = {}

    def add(self, name: str, timeout_override: int = None) -> int:
        if not name in self.buckets:
            self.buckets[name] = LeakyBucket(timeout_override if timeout_override else self.timeout)

        self.buckets[name].add()
        return self.buckets[name].get()

    def get(self, name: str) -> int:
        if not name in self.buckets:
            return 0

        return self.buckets[name].get()